/*
 Copyright (C) 1997-2001 Id Software, Inc.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

#include <termios.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <signal.h>
#include <sys/mman.h>

#if defined (__linux__)
#include <asm/io.h>
#include <sys/vt.h>
#endif

#include "../ref_soft/r_local.h"
#include "../client/keys.h"
#include "rw_android.h"
#include "quake2-jni.h"

#include <android/log.h>

#include "sensor_input.h"



#define LOGI(...) ((void)__android_log_print(ANDROID_LOG_INFO, "native-activity", __VA_ARGS__))

#include "udpsocket.h"

static cvar_t  *cl_forwardspeed;

float movey,movex;


struct UDPSensorPacket{
	char version;
	char command;
	struct timeval timestamp;
	float yaw;
	float pitch;
	float roll;
	char contains_move_commands;
	float move_x;
	float move_y;
	char eightbuttons;
};


#define UDP_COMMAND_NOCOMMAND 0
#define UDP_COMMAND_REQUESTSENSOR 1
#define UDP_COMMAND_SENDSENSOR 2
#define UDP_COMMAND_SENDMOVE 3



int IsBitSet(char b, int pos)
{
   return (b & (1 << pos)) != 0;
}

int UDPSensorPacket_deserialize(struct UDPSensorPacket * p,char * buf){

	 int i=0;
	 char *tok = strtok(buf, " ");
	 p->version= atoi(tok);tok = strtok(NULL, " ");
	 p->command= atoi(tok);tok = strtok(NULL, " ");
	 p->timestamp.tv_sec= atoi(tok);tok = strtok(NULL, " ");

	 p->timestamp.tv_usec= atoi(tok);tok = strtok(NULL, " ");
	 p->yaw=atof(tok);tok=strtok(NULL," ");
	 p->pitch=atof(tok);tok=strtok(NULL," ");
	 p->roll=atof(tok);tok=strtok(NULL," ");
	 p->contains_move_commands=atoi(tok);  tok=strtok(NULL," ");
	 p->move_x=atof(tok);  tok=strtok(NULL," ");
	 p->move_y=atof(tok);  tok=strtok(NULL," ");
	 p->eightbuttons=atoi(tok);


}







/*****************************************************************************/
/* KEYBOARD                                                                  */
/*****************************************************************************/

static Key_Event_fp_t Key_Event_fp;





#define MAX_KEY_EVENTS 128

static int keyEvents[MAX_KEY_EVENTS];
static int eventId ;

struct UDPsocket sock;
#define MAXMESG 1024
char buf[MAXMESG];
char dummybuf[MAXMESG];
 int len;
 int dummylen;
float float_array[10];
struct UDPSensorPacket udpsensorpacket;


// isReadable
//   Checks to see if the Socket has something to be read
int isReadable( int sockfd, int usec ) {
  fd_set socketReadSet;
  FD_ZERO(&socketReadSet);
  FD_SET(sockfd, &socketReadSet);

  // Setup timeout
  struct timeval tv;
  if( usec ) {
    tv.tv_sec   = usec / 1000000;
    tv.tv_usec = (usec % 1000000);
  } else {
    tv.tv_sec  = 0;
    tv.tv_usec = 0;
  } // end if



  // Select on our Socket Description
  if( select(sockfd+1,&socketReadSet,0,0,&tv) == -1 ) {
    perror("select");
    exit(1);
  }

  return FD_ISSET(sockfd,&socketReadSet) != 0;

} // end isReadable


void KBD_Init(Key_Event_fp_t fp) {
    //__android_log_print(ANDROID_LOG_DEBUG, "KBD", "KBD_Init\n");
    Key_Event_fp = fp;
    eventId = 0;
}

        
void KBD_Update(void) {
	
	int k;
	
    //__android_log_print(ANDROID_LOG_DEBUG, "rw_in_android.c", "KBD_Update\n");
	
	for (k=0;k<eventId;k++){
		int event = keyEvents[k];
	                	           
	    int key = event & 0x00ff;
	    int down = (event>>8) & 0x00ff;
	                	
		Key_Event_fp(key, down);
	}
	
	eventId = 0;
}



void quake2_jni_key_event( int key, int down) 
{

	if ( eventId < MAX_KEY_EVENTS )
		keyEvents[eventId++] = key | down<<8;

}


void KBD_Close(void) {

}

/*****************************************************************************/
/* MOUSE                                                                     */
/*****************************************************************************/


static in_state_t *in_state;


static struct {
	int mode;
	int forwardmove, sidemove, upmove;
	float pitch, yaw, roll;
} event;


	
// this is inside the renderer shared lib, so these are called from vid_so


void RW_IN_Init(in_state_t *in_state_p) {
	

    cl_forwardspeed = ri.Cvar_Get ("cl_forwardspeed", "200", 0);


	in_state = in_state_p;

	event.mode = 0;
	// INIT the UDP Receive Socket
		 //         //
    __android_log_print(ANDROID_LOG_DEBUG, "UDP", "INITING SOCKET\n");

//	  LOGI("\ninit_socket_fd = ");
	  UDPsocket_init_socket_fd(&sock,0);

	    __android_log_print(ANDROID_LOG_DEBUG, "UDP", "set server adress\n");

	  //LOGI("\ninit_serv_addr = ");
	  UDPsocket_init_serv_addr(&sock,"localhost",12345);
	 // UDPsocket_set_fd_nonblock(&sock);

	  __android_log_print(ANDROID_LOG_DEBUG, "UDP", "send message:");
	  //LOGI("\nsend_msg       = ");
	  //UDPsocket_send_msg(&sock,"UDP_PROCESS",11);
	  //__android_log_print(ANDROID_LOG_DEBUG, "UDP", "receive message\n");
	  //LOGI("\nrecv_msg       = ");
	  //UDPsocket_recv_msg(&sock,buf,len,4096,1);
	  //__android_log_print(ANDROID_LOG_DEBUG, "UDP", "receive message\n");

	 // int UDPsocket_recv_msg(struct UDPsocket*s,char *buf, int  len, int max_len, int redirect) {

	  //LOGI("\nbuf= %s",buf);
	  //UDPsocket_recv_msg(&sock,"(dash 100)",10,4096,1);
	  //__android_log_print(ANDROID_LOG_DEBUG, "UDP", "close socket\n");
//	  UDPsocket_close_socket_fd(&sock);





	//init_sensor();


    //__android_log_print(ANDROID_LOG_DEBUG, "KBD", "RW_IN_Init\n");
}

void RW_IN_Shutdown(void) {
    //__android_log_print(ANDROID_LOG_DEBUG, "KBD", "RW_IN_Shutdown\n");
}

/*
 ===========
 IN_Commands
 ===========
 */
void RW_IN_Commands(void) {
    //__android_log_print(ANDROID_LOG_DEBUG, "KBD", "RW_IN_Commands\n");
}

/*
 ===========
 IN_Move
 ===========
 */




	
void quake2_jni_move_event( int mode, 
				int forwardmove, int sidemove, int upmove,
				float pitch, float yaw, float roll
				) {
	
	/*
	__android_log_print(ANDROID_LOG_DEBUG, "rw_in_android.c", 
					"quake2_jni_move_event forwardmove= %d yaw= %.0f\n", forwardmove, yaw);
*/
	event.mode = mode;
	
	event.forwardmove = forwardmove;
	event.sidemove = sidemove;
	event.upmove = upmove;

	
	event.pitch = pitch;
	event.yaw = yaw;
	event.roll = roll;	
	
	
	
}


void RW_IN_Move(usercmd_t *cmd) {
	
	  struct timeval start,current;

	    long mtime, seconds, useconds;


    __android_log_print(ANDROID_LOG_DEBUG, "rw_in_android.c", "RW_IN_Move\n");

	if ( event.mode == 1 ){ // delta mode
	/*
		__android_log_print(ANDROID_LOG_DEBUG, "rw_in_android.c", 
				"RW_IN_Move forwardmove= %d yaw= %.0f\n", event.forwardmove, event.yaw);
		*/
		
		cmd->forwardmove += event.forwardmove;
		cmd->sidemove += event.sidemove;
		cmd->upmove += event.upmove;
		
		in_state->viewangles[PITCH] += event.pitch;
		in_state->viewangles[YAW] += event.yaw ;
		in_state->viewangles[ROLL] += event.roll ;

	} else if ( event.mode == 2 ){
		
		cmd->forwardmove += event.forwardmove;
		cmd->sidemove += event.sidemove;
		cmd->upmove += event.upmove;
		
		in_state->viewangles[PITCH] = event.pitch;
		in_state->viewangles[YAW] += event.yaw ;
		in_state->viewangles[ROLL] += event.roll ;
	} else if ( event.mode == 3 ){
		
		cmd->forwardmove += event.forwardmove;
		cmd->sidemove += event.sidemove;
		cmd->upmove += event.upmove;
		
		in_state->viewangles[PITCH] = event.pitch;
		in_state->viewangles[YAW] = event.yaw ;
		in_state->viewangles[ROLL] = event.roll ;
	}
	
	


    gettimeofday(&start, NULL);


     LOGI("asking for UDP Message at %f   \n",start.tv_usec*1.0f/1000.0f);
    gettimeofday(&start, NULL);
 	     gettimeofday(&current, NULL);



	  UDPsocket_send_msg(&sock,"1 1\n",4);
	  LOGI("waiting for UDP Message at %f   \n",start.tv_usec*1.0f/1000.0f);

	  int isreadable=true;
	  int last_readable_len;

		/* isreadable = UDPsocket_recv_msg(&sock,buf,&len,MAXMESG,1);
//
		 buf[len]="\0";
				 int i=0;
				 char *tok = strtok(buf, " ");
				 while (tok != NULL) {
				 float_array[i++] = atof(tok);
				 tok = strtok(NULL, " ");
				 }
*/
	     LOGI("going to block .");

	  if (  isReadable(sock.socket_fd,10000))
	  {
		  LOGI("In select");
		  len=1;
		  isreadable = UDPsocket_recv_msg(&sock,buf,&len,MAXMESG,1);
		  				 LOGI ("length of receive buffer  is is %i",len);
		  				 if (len > 0){
		  					 buf[len]="\0";

		  					 UDPSensorPacket_deserialize(&udpsensorpacket,buf);


		  				 }





		  //while (len >0){
			  while (isReadable(sock.socket_fd,0)){
				 isreadable = UDPsocket_recv_msg(&sock,buf,&len,MAXMESG,1);
				 LOGI ("ADDITIONAL PACKETS length of receive buffer  is is %i",len);
				 //LOGI ("received message %s",buf);
				 if (len > 0){
					 buf[len]="\0";
					 UDPSensorPacket_deserialize(&udpsensorpacket,buf);

				 }


				 }



	  }

	  //LOGI ("received %i",len);

	  // received something{



	  char buttons;

	  if (udpsensorpacket.contains_move_commands){
		  movex=udpsensorpacket.move_x* cl_forwardspeed->value;
		  movey=udpsensorpacket.move_y* cl_forwardspeed->value;
		  buttons=udpsensorpacket.eightbuttons;
		  LOGI("cONTAINS MOVECOMMANDS !!!!!!! %i",buttons);
		  if (IsBitSet(buttons, 0)){
			  	  	  	  	  	  	  LOGI("Button 0 pressed");
			  	  	  	  	  	  	  quake2_jni_key_event("a",1);
		  	  	  }

		  quake2_jni_key_event(112,IsBitSet(buttons, 0));
		  quake2_jni_key_event(27,IsBitSet(buttons, 1));
		  quake2_jni_key_event(13,IsBitSet(buttons, 2));
		  quake2_jni_key_event(99,IsBitSet(buttons, 3));
		  quake2_jni_key_event(106,IsBitSet(buttons, 4));


	  }


	  cmd->forwardmove+=movey;
	  cmd->sidemove+=movex;

	            gettimeofday(&current, NULL);
       // 		LOGI("received  at %f   \n",start.tv_usec*1.0f/1000.0f);


	__android_log_print(ANDROID_LOG_DEBUG, "UDP","I have received y %f p %f r=%f time=%i",udpsensorpacket.yaw,udpsensorpacket.pitch,udpsensorpacket.roll,(udpsensorpacket.timestamp.tv_usec-current.tv_usec)/1000);
    LOGI(" Difference in time =  time=%f",(udpsensorpacket.timestamp.tv_usec-current.tv_usec)*1.0f/1000);
	in_state->viewangles[PITCH]=udpsensorpacket.pitch;
	in_state->viewangles[ROLL]=udpsensorpacket.roll;
	in_state->viewangles[YAW]=udpsensorpacket.yaw;


if (udpsensorpacket.contains_move_commands){
	LOGI ("I found some move commands !!!!!!!");

}

	event.mode = 0; // no event 
	
	

}

void RW_IN_Frame(void) {
    //__android_log_print(ANDROID_LOG_DEBUG, "KBD", "RW_IN_Frame\n");
}

void RW_IN_Activate(void) {
    //__android_log_print(ANDROID_LOG_DEBUG, "KBD", "RW_IN_Activate\n");
}

